# -*- coding: utf8 -*-

import math
from functools import cmp_to_key

from driverng import JSONGraphContainer


DOWN, UP = range(2)


def get_epaulements(points):
    atan = [math.atan(points[i] - points[i - 1]) for i in range(1, len(points))]
    epaulements = []
    for i in range(1, len(atan)):
        diff = atan[i] - atan[i - 1]
        if diff > 0:
            direction = UP
        else:
            direction = DOWN
        if not epaulements or epaulements[-1]["direction"] != direction:
            epaulements.append({"direction": direction, "x": i, "y": atan[i]})
    return epaulements


def get_graph_analysis(graph):
    sa1c, a0 = None, None
    for peak in graph["peaks"]:
        if peak["name"] == "SA1C":
            sa1c = peak
        if peak["name"] == "A0":
            a0 = peak

    results = []

    if sa1c:
        # Épaulement gauche
        ep_sa1c_left = 0
        epaulements = get_epaulements(graph["points"][sa1c["start"] : sa1c["top"]])
        for i in range(len(epaulements)):
            if [e["direction"] for e in epaulements[i : i + 3]] == [DOWN, UP, DOWN]:
                ep_sa1c_left = max(
                    ep_sa1c_left,
                    abs(
                        epaulements[i + 1]["y"]
                        - min(epaulements[i]["y"], epaulements[i + 2]["y"])
                    ),
                )

        # Épaulement droit
        ep_sa1c_right = 0
        epaulements = get_epaulements(graph["points"][sa1c["top"] : sa1c["end"]])
        for i in range(len(epaulements)):
            if [e["direction"] for e in epaulements[i : i + 3]] == [UP, DOWN, UP]:
                ep_sa1c_right = max(
                    ep_sa1c_right,
                    abs(
                        epaulements[i + 1]["y"]
                        - max(epaulements[i]["y"], epaulements[i + 2]["y"])
                    ),
                )

        results.extend(
            [
                {
                    "code": "A1C_EPA1G",
                    "value": str(round(math.degrees(ep_sa1c_left), 2)),
                },
                {
                    "code": "A1C_EPA1D",
                    "value": str(round(math.degrees(ep_sa1c_right), 2)),
                },
            ]
        )

        # Hauteur en pourcentage du début SA1C par rapport à la
        # hauteur min du début du pic précédent jusqu'à la fin du
        # SA1C.
        # Permet de voir si le début du SA1C est un ceux bien marqué.
        peaks = sorted(
            graph["peaks"], key=cmp_to_key(lambda a, b: a["start"] - b["start"])
        )
        names = [p["name"] for p in peaks]
        before_sa1c_index = names.index("SA1C") - 1
        if before_sa1c_index >= 0:
            peak = peaks[before_sa1c_index]
            vmin = min(graph["points"])
            points = [v - vmin for v in graph["points"]]
            height = points[sa1c["start"]]
            height_p = (height - min(points[peak["start"] : sa1c["end"]])) / height
            results.append(
                {
                    "code": "A1C_START_SA1C_HEIGHT_P",
                    "value": str(round(height_p * 100, 2)),
                }
            )

    if sa1c and a0:
        # Aire en pourcentage par rapport à l'aire totale.
        # de la fin de SA1C jusqu'au dernier point avant le top A0 ou
        # la courbe a une tangente avec un angle inférieur à 45°.
        vmin = min(graph["points"])
        points = [v - vmin for v in graph["points"]]
        part = points[sa1c["end"] : a0["top"]]
        last = None
        for i in range(1, len(part)):
            if part[i] - part[i - 1] < 1:  # atan(1) = 45°
                last = sa1c["end"] + i

        total = sum(points)
        ep_a0_area = sum(points[sa1c["end"] : last])

        results.extend(
            [{"code": "A1C_EPA0_P", "value": str(round((ep_a0_area / total) * 100, 2))}]
        )

    return results


def get_measurement_values(self, message):
    mode = message.sample_information.operation_mode
    values = [c.value for c in message.measurement_value.components]
    values.append(message.measurement_value.reserved)
    results = []
    for index, value in enumerate(values):
        try:
            # Version G11 utilisant le mode Standard (mode 0) ou le mode Variant (mode 1) ou le mode
            # Beta-thalassemia (mode 2) par autodétection du mode dans la ligne de
            # type 1
            if self.version_tosoh == "version_g11":
                test = self.COMPONENT_MAP_CDS_G11[mode][index]
            else:
                test = self.COMPONENT_MAP[mode][index]
        except IndexError:
            continue
        if not test:
            continue
        results.append((test, value))
    return results


def get_graphe(self, message):
    g = {"graph_type": "a1c", "points": [], "peaks": []}
    cp = message.data_collection_parameters
    g["parameters"] = {
        "start_time": int(cp.collection_start_time),
        "pitch": int(cp.collection_pitch),
        "n_peaks": int(cp.number_of_peaks),
    }
    for raw in message.raw_datas:
        # L'automate GX complète la trame avec 9 points avec des valeurs 0.000
        # Ces points de valeur 0.000 ne sont pas à prendre en compte, il
        # faut poursuivre la linéarité
        if (
            self.version_tosoh == "version_gx"
            and len(set([val.value for val in raw.block_signals[-9:]]) - set(["0.000"]))
            == 0
        ):
            raw.block_signals[-9].value = raw.block_signals[-10].value
            raw.block_signals[-8].value = raw.block_signals[-10].value
            raw.block_signals[-7].value = raw.block_signals[-10].value
            raw.block_signals[-6].value = raw.block_signals[-10].value
            raw.block_signals[-5].value = raw.block_signals[-10].value
            raw.block_signals[-4].value = raw.block_signals[-10].value
            raw.block_signals[-3].value = raw.block_signals[-10].value
            raw.block_signals[-2].value = raw.block_signals[-10].value
            raw.block_signals[-1].value = raw.block_signals[-10].value
        # L'automate G11 avec le protocole "CS" se rapprochant au protocole
        # G8 mode raw complète la trame avec 2 points avec des valeurs 0.000
        # Ces points de valeur 0.000 ne sont pas à prendre en compte, il
        # faut poursuivre la linéarité (vu avec TOSOH le 15/11/2016 chez
        # BIOPATH)
        elif (
            raw.block_signals[-2].value == "0.000"
            and raw.block_signals[-1].value == "0.000"
        ):
            raw.block_signals[-2].value = raw.block_signals[-3].value
            raw.block_signals[-1].value = raw.block_signals[-3].value
        for point in raw.block_signals:
            v = float(point.value)
            g["points"].append(v)
    for peak in message.peak_datas:
        g["peaks"].append(
            {
                "name": peak.name,
                "shape": peak.shape,
                "base_start_point": int(peak.base_start_point),
                "base_end_point": int(peak.base_end_point),
                "start": int(peak.peak_start_point),
                "top": int(peak.peak_top),
                "end": int(peak.peak_end_point),
                "area": float(peak.peak_area),
                "area_percent": float(peak.peak_area_percent),
            }
        )
    results = get_graph_analysis(g)
    results.append(
        {"code": "GRAPHE_A1C", "value": "GRAPHE", "graph": JSONGraphContainer(g)}
    )
    return results


def get_legacy_graphe(self, message, alarmes):
    resu_tube = []
    cp = message.data_collection_parameters
    resu_tube.append({"code": "NbPick", "value": cp.number_of_peaks, "flags": alarmes})
    aireTotal = sum([float(p.peak_area) for p in message.peak_datas])
    resu_tube.append(
        {"code": "areaTotal", "value": "{:.2f}".format(aireTotal), "flags": alarmes}
    )
    pourcTotal = sum([float(p.peak_area_percent) for p in message.peak_datas])
    resu_tube.append(
        {"code": "pourcTotal", "value": "{:.2f}".format(pourcTotal), "flags": alarmes}
    )

    tempsTotal = 0
    for peak in message.peak_datas:
        tr = (
            float(cp.collection_start_time)
            + (int(peak.peak_top) * int(cp.collection_pitch))
        ) / 600
        tempsTotal += tr
        if self.version_tosoh == "version_g11":
            resu = "{:.4f}".format(tr / 2)
        else:
            resu = "{:.4f}".format(tr)
        resu_tube.append(
            {"code": "tpr_{:s}".format(peak.name), "value": resu, "flags": alarmes}
        )
        resu_tube.append(
            {
                "code": "area_{:s}".format(peak.name),
                "value": "{:.2f}".format(float(peak.peak_area)),
                "flags": alarmes,
            }
        )
        resu_tube.append(
            {
                "code": "areap_{:s}".format(peak.name),
                "value": "{:.2f}".format(float(peak.peak_area_percent)),
                "flags": alarmes,
            }
        )

    resu_tube.append(
        {"code": "tempsTotal", "value": "{:.3f}".format(tempsTotal), "flags": alarmes}
    )
    return resu_tube


def get_sample_id(self, message):
    if self.sample_id_mode == "ignore_2":
        sample_id = message.sample_information.sample_id[2:]
    elif self.sample_id_mode == "normal":
        sample_id = message.sample_information.sample_id
    else:
        raise NotImplementedError("sample_id mode {:s}".format(self.sample_id_mode))

    return sample_id


class One(object):
    def __init__(self, name, klass):
        self.name = name
        self.klass = klass

    def get(self, **kwargs):
        return self.klass(**kwargs)


class ManyContainer(list):
    def __init__(self, klass):
        self.klass = klass
        list.__init__(self)

    def add(self, **kwargs):
        obj = self.klass(**kwargs)
        self.append(obj)
        return obj


class Many(One):
    def get(self):
        return ManyContainer(self.klass)
